
#include "main.h"
#include "tex_area.h"
#include "misc.h"

extern Wad iwad;

TextureArea::TextureArea(wxWindow *parent, wxColour *palette, Wad *wadfile)
:	wxPanel(parent, -1, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER)
{
	SetBackgroundColour(wxColour(50, 50, 50));
	this->palette = palette;
	this->wadfile = wadfile;
	texture = NULL;
	sel_patch = -1;
}

TextureArea::~TextureArea()
{
}

void TextureArea::setTexture(texture_t *tex)
{
	if (texture == tex)
		return;

	for (int a = 0; a < images.size(); a++)
		delete images[a];

	images.clear();

	texture = tex;

	if (!tex)
		return;

	for (int a = 0; a < tex->patches.size(); a++)
	{
		Wad* wad = wadfile;
		wxImage *image = new wxImage();
		Lump* lump = wadfile->getLump(tex->patches[a].patch, 0);

		if (!lump)
		{
			lump = iwad.getLump(tex->patches[a].patch, 0);
			wad = &iwad;
		}

		wad->getLump(tex->patches[a].patch, 0)->determineType();

		image_from_lump(lump, image, palette);
		images.push_back(image);
	}
}

BEGIN_EVENT_TABLE(TextureArea, wxPanel)
	EVT_PAINT(TextureArea::paint)
END_EVENT_TABLE()

void TextureArea::paint(wxPaintEvent &event)
{
	wxPaintDC dc(this);

	dc.BeginDrawing();

	if (texture)
	{
		wxPoint topleft((GetClientSize().x / 2) - (texture->width / 2), (GetClientSize().y / 2) - (texture->height / 2));

		for (int a = 0; a < texture->patches.size(); a++)
			dc.DrawBitmap(wxBitmap(*images[a]), topleft.x + texture->patches[a].xoff, topleft.y + texture->patches[a].yoff, true);

		dc.SetBrush(wxBrush(wxColour(), wxTRANSPARENT));
		dc.SetPen(*wxWHITE_PEN);
		dc.DrawRectangle(topleft.x, topleft.y, texture->width, texture->height);

		if (sel_patch >= 0 && sel_patch < texture->patches.size())
		{
			dc.SetBrush(wxBrush(wxColour(), wxTRANSPARENT));
			dc.SetPen(*wxRED_PEN);
			dc.DrawRectangle(topleft.x + texture->patches[sel_patch].xoff, topleft.y + texture->patches[sel_patch].yoff,
				images[sel_patch]->GetWidth(), images[sel_patch]->GetHeight());
		}
	}

	dc.EndDrawing();
}

TextureLumpArea::TextureLumpArea(wxWindow *parent, wxColour* palette, Wad* wadfile)
:	LumpArea(parent)
{
	type = LAREA_TEXTURES;
	this->wadfile = wadfile;

	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(hbox);

	// Textures frame
	wxStaticBox *frame = new wxStaticBox(this, -1, "Textures");
	wxStaticBoxSizer *box = new wxStaticBoxSizer(frame, wxVERTICAL);
	hbox->Add(box, 0, wxEXPAND|wxALL, 4);

	list_textures = new wxListBox(this, TLA_LIST_TEXTURES, wxDefaultPosition, wxSize(96, -1));
	box->Add(list_textures, 1, wxEXPAND|wxALL, 4);

	// Texture draw area
	ta_texture = new TextureArea(this, palette, wadfile);
	hbox->Add(ta_texture, 1, wxEXPAND|wxALL, 4);

	wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);
	hbox->Add(vbox, 0, wxEXPAND);

	// PNames frame
	frame = new wxStaticBox(this, -1, "PNames");
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	vbox->Add(box, 2, wxEXPAND|wxALL, 4);

	list_pnames = new wxListBox(this, -1, wxDefaultPosition, wxSize(96, -1));
	box->Add(list_pnames, 1, wxEXPAND|wxALL, 4);

	// Read PNames
	pnames.clear();
	DWORD n_pnames = 0;
	Wad* wad = wadfile;
	Lump *lump = wad->getLump("PNAMES", 0);

	if (!lump)
	{
		lump = iwad.getLump("PNAMES", 0);
		wad = &iwad;
	}

	if (lump)
	{
		FILE *fp = fopen(wad->path.c_str(), "rb");
		fseek(fp, lump->getOffset(), SEEK_SET);
		fread(&n_pnames, 4, 1, fp);

		for (DWORD p = 0; p < n_pnames; p++)
		{
			char name[9] = "";
			fread(name, 1, 8, fp);
			name[8] = 0;
			pnames.push_back(name);
			list_pnames->Append(name);
		}

		fclose(fp);
	}

	// Patches frame
	frame = new wxStaticBox(this, -1, "Patches");
	box = new wxStaticBoxSizer(frame, wxVERTICAL);
	vbox->Add(box, 1, wxEXPAND|wxALL, 4);

	list_patches = new wxListBox(this, TLA_LIST_PATCHES, wxDefaultPosition, wxSize(96, -1));
	box->Add(list_patches, 1, wxEXPAND|wxALL, 4);

	Layout();
	ta_texture->Refresh();
	ta_texture->Update();
}

TextureLumpArea::~TextureLumpArea()
{
}

void TextureLumpArea::loadLump(Lump *lump)
{
	long n_tex = 0;
	long *offsets = NULL;
	long temp = 0;

	short width = 0;
	short height = 0;
	short patches = 0;

	short xoff = 0;
	short yoff = 0;
	short patch = 0;

	list_textures->Clear();

	FILE *fp = fopen(wadfile->path.c_str(), "rb");
	fseek(fp, lump->getOffset(), SEEK_SET);

	// Get no. of textures and tex info offsets
	fread(&n_tex, 4, 1, fp);
	offsets = (long *)calloc(n_tex, 4);
	fread(offsets, 4, n_tex, fp);

	for (int t = 0; t < n_tex; t++)
	{
		// Go to start of tex definition
		fseek(fp, lump->getOffset(), SEEK_SET);
		fseek(fp, offsets[t], SEEK_CUR);

		// Read texture name
		char texname[9] = "";
		fread(texname, 1, 8, fp);
		texname[8] = 0;

		// Skip unused stuff
		fread(&temp, 4, 1, fp);

		// Read width & height
		fread(&width, 2, 1, fp);
		fread(&height, 2, 1, fp);

		// Add texture
		texture_t tex;
		tex.name = texname;
		tex.width = width;
		tex.height = height;
		list_textures->Append(tex.name);

		// Skip more unused stuff
		fread(&temp, 4, 1, fp);

		// Read no. of patches in texture
		fread(&patches, 2, 1, fp);

		// Add patches
		for (int p = 0; p < patches; p++)
		{
			// Read patch info
			fread(&xoff, 2, 1, fp);
			fread(&yoff, 2, 1, fp);
			fread(&patch, 2, 1, fp);

			// Skip unused
			fread(&temp, 4, 1, fp);

			patch_info_t pi;
			pi.patch = pnames[patch];
			pi.xoff = xoff;
			pi.yoff = yoff;
			tex.patches.push_back(pi);
		}

		textures.push_back(tex);
	}

	fclose(fp);
}

void TextureLumpArea::checkSave()
{
}

BEGIN_EVENT_TABLE(TextureLumpArea, wxPanel)
	EVT_LISTBOX(TLA_LIST_TEXTURES, TextureLumpArea::onTextureListChange)
	EVT_LISTBOX(TLA_LIST_PATCHES, TextureLumpArea::onPatchesListChange)
END_EVENT_TABLE()

void TextureLumpArea::onTextureListChange(wxCommandEvent &event)
{
	list_patches->Clear();

	ta_texture->setTexture(&textures[list_textures->GetSelection()]);
	ta_texture->selPatch(-1);
	ta_texture->Refresh();
	ta_texture->Update();

	texture_t tex = textures[list_textures->GetSelection()];
	for (int a = 0; a < tex.patches.size(); a++)
		list_patches->Append(tex.patches[a].patch);
		//list_patches->Append(wxString::Format("%s %d %d", tex.patches[a].patch, tex.patches[a].xoff, tex.patches[a].yoff));
}

void TextureLumpArea::onPatchesListChange(wxCommandEvent &event)
{
	ta_texture->selPatch(list_patches->GetSelection());
	ta_texture->Refresh();
	ta_texture->Update();
}

