////////////////////////////////////////////////////////////////////////
// SLump - A doom wad/lump manager                                    //
// By Simon Judd, 2006                                                //
// ------------------------------------------------------------------ //
// wad_panel.cpp - WadPanel class functions                           //
////////////////////////////////////////////////////////////////////////

#include "main.h"
#include "wad_panel.h"
#include "main_window.h"
#include "gfx_area.h"
#include "tex_area.h"
#include "map_area.h"
#include "text_area.h"

#include <wx/filename.h>

vector<WadPanel*> open_wad_panels;

extern MainWindow *main_window;
extern Wad iwad;

/*
struct dirinfo_t
{
	string			name;
	int				level;
	wxTreeItemId	id;

	dirinfo_t(string iname, int ilevel, wxTreeItemId iid)
	{
		name = iname;
		level = ilevel;
		id = iid;
	}
};
vector<dirinfo_t> dir_struct;
*/

WadPanel::WadPanel(wxWindow *parent, string wadpath, bool zip)
:	wxPanel(parent, -1)
{
	// Load the wad file
	opened_ok = false;
	open_wad_panels.push_back(this);

	if (wadpath != _T(""))
	{
		string ext = wadpath.Right(3);

		if (ext == _T("wad"))
		{
			// If the wadfile failed to open, return
			if (!wadfile.open(wadpath, true))
				return;
		}

		if (ext == _T("zip") || ext == _T("pk3"))
		{
			// If the wadfile failed to open, return
			if (!wadfile.openZip(wadpath, true))
				return;
		}
	}
	else
		wadfile.zip = zip;

	// Wadfile opened ok
	opened_ok = true;
	main_window->enableWadMenus(true);

	// Setup palette
	palette_path = _T("res/palettes/doom.pal");
	loadPalette();

	wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(hbox);

	wxStaticBox *frame = new wxStaticBox(this, -1, _T("Lumps"));
	wxStaticBoxSizer *framesizer = new wxStaticBoxSizer(frame, wxVERTICAL);

	lump_list = new wxListCtrl(this, WP_LIST_LUMPS, wxDefaultPosition, wxDefaultSize, wxLC_REPORT/*|wxLC_EDIT_LABELS*/);
	lump_list->SetFont(wxFont(10, wxFONTFAMILY_MODERN, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL));
	framesizer->Add(lump_list, 1, wxEXPAND);
	hbox->Add(framesizer, 0, wxEXPAND|wxALL, 4);

	hbox->AddSpacer(4);

	frame = new wxStaticBox(this, -1, _T("Lump Content"));
	lump_area_sizer = new wxStaticBoxSizer(frame, wxVERTICAL);

	lump_area = new LumpArea(this);
	lump_area_sizer->Add(lump_area, 1, wxEXPAND);
	hbox->Add(lump_area_sizer, 1, wxEXPAND|wxALL, 4);

	lump_list->InsertColumn(0, _T("Name"));
	lump_list->InsertColumn(1, _T("Size"), wxLIST_FORMAT_RIGHT);
	lump_list->InsertColumn(2, _T("Type"), wxLIST_FORMAT_RIGHT);
	for (int a = 0; a < wadfile.numLumps(); a++)
	{
		wxListItem li;

		li.SetText(wadfile.lumpAt(a)->getName());
		li.SetId(a);
		li.SetColumn(0);
		li.SetData(wadfile.lumpAt(a));
		lump_list->InsertItem(li);

		li.SetText(wxString::Format(_T("%d"), wadfile.lumpAt(a)->getSize()));
		li.SetColumn(1);
		li.SetId(a);
		lump_list->SetItem(li);

		li.SetText(wadfile.lumpAt(a)->getTypeString());
		li.SetColumn(2);
		lump_list->SetItem(li);
	}

	int width = 0;
	for (int a = 0; a < lump_list->GetColumnCount(); a++)
		width += lump_list->GetColumnWidth(a);

	lump_list->SetSizeHints(width + 20, -1, width + 20, -1);

	Layout();
	parent->Layout();
	Refresh();
	parent->Refresh();
}

WadPanel::~WadPanel()
{
	open_wad_panels.erase(find(open_wad_panels.begin(), open_wad_panels.end(), this));
}

string WadPanel::getFilename()
{
	if (wadfile.path == _T(""))
		return _T("UNSAVED");
	else
		return wadfile.path;
}

void WadPanel::loadPalette()
{
	Lump* lump = wadfile.getLump("PLAYPAL", 0);

	if (!lump)
		lump = iwad.getLump("PLAYPAL", 0);

	for (int a = 0; a < 256; a++)
	{
		if (lump)
			palette[a].Set(lump->getData()[a*3], lump->getData()[(a*3)+1], lump->getData()[(a*3)+2]);
		else
			palette[a].Set(a, a, a);
	}
}

bool WadPanel::closeWad()
{
	lump_area->checkSave();

	if (wadfile.needsSave())
	{
		string question;
		if (wadfile.path == _T(""))
			question = _T("Save Changes to the wad?");
		else
			question = wxString::Format(_T("Save changes to \"%s\"?"), wadfile.path);

		int ret = wxMessageBox(question, _T("Close Wad"), wxICON_QUESTION|wxYES_NO|wxCANCEL);

		if (ret == wxCANCEL)
			return false;

		if (ret == wxYES)
			saveWad();
	}

	return true;
}

void WadPanel::saveWad()
{
	if (wadfile.path == _T(""))
		saveWadAs();
	else
		wadfile.save();
}

void WadPanel::saveWadAs()
{
	wxString filename = wxFileSelector(_T("Save Wad"), _T(""), _T(""), _T("*.wad"), _T("Doom Wad Files (*.wad)|*.wad"),
										wxSAVE|wxOVERWRITE_PROMPT);

	if (!filename.empty())
	{
		wadfile.path = filename;
		wadfile.save();
	}
}

void WadPanel::addLumpToList(int index, Lump *lump)
{
	wxListItem li;
	li.SetId(index);
	li.SetText(lump->getName());
	li.SetData(lump);
	lump_list->InsertItem(li);
	li.SetText(wxString::Format(_T("%d"), lump->getSize()));
	li.SetColumn(1);
	lump_list->SetItem(li);
	li.SetText(lump->getTypeString());
	li.SetColumn(2);
	lump_list->SetItem(li);

	updateSelection();
}

void WadPanel::updateSelection()
{
	selection.clear();

	int s = 0;
	for (int a = 0; a < lump_list->GetItemCount(); a++)
	{
		if (lump_list->GetItemState(a, wxLIST_STATE_SELECTED))
		{
			selection.push_back(a);
			s++;
		}

		if (s == lump_list->GetSelectedItemCount())
			return;
	}
}

void WadPanel::updateList(int index)
{
	if (index < 0 || index >= lump_list->GetItemCount())
		return;

	Lump* lump = (Lump*)lump_list->GetItemData(index);

	wxListItem li;
	li.SetId(index);
	li.SetText(lump->getName());
	lump_list->SetItem(li);
	li.SetText(wxString::Format(_T("%d"), lump->getSize()));
	li.SetColumn(1);
	lump_list->SetItem(li);
	li.SetText(lump->getTypeString());
	li.SetColumn(2);
	lump_list->SetItem(li);
}

void WadPanel::updateList(Lump* lump)
{
	for (int a = 0; a < lump_list->GetItemCount(); a++)
	{
		Lump* tlump = (Lump*)lump_list->GetItemData(a);

		if (lump == tlump)
		{
			updateList(a);
			return;
		}
	}
}

void WadPanel::newLump()
{
	int index = 0;
	if (selection.size() > 0)
		index = selection.back() + 1;

	wxString name = wxGetTextFromUser(_T("Enter new lump name:"), _T("New Lump"));

	if (!wadfile.zip)
	{
		name.UpperCase();
		name.Truncate(8);
	}

	Lump *nlump = wadfile.addLump(name, index);
	addLumpToList(index, nlump);
}

void WadPanel::newLumpFromFile()
{
	int index = 0;
	if (selection.size() > 0)
		index = selection.back() + 1;

	wxFileDialog *OpenDialog = new wxFileDialog(this, _T("Choose file(s) to open"), wxEmptyString, wxEmptyString, 
		_T("All files (*.*)|*.*"), wxOPEN|wxMULTIPLE|wxFILE_MUST_EXIST, wxDefaultPosition);

	if (OpenDialog->ShowModal() == wxID_OK)
	{
		wxArrayString files;
		OpenDialog->GetPaths(files);

		for (int a = 0; a < files.size(); a++)
		{
			wxFileName name(files[a]);
			Lump *nlump = wadfile.addLump(name.GetName().Truncate(8).Upper(), index);
			nlump->loadFile(files[a]);
			nlump->determineType();
			addLumpToList(index, nlump);
		}
	}
}

void WadPanel::deleteLump()
{
	while (selection.size() > 0)
	{
		int index = selection[0];
		wadfile.deleteLump(index);
		lump_list->DeleteItem(index);
		updateSelection();
	}
}

void WadPanel::renameLump()
{
	if (selection.size() > 0)
	{
		for (int a = 0; a < selection.size(); a++)
		{
			Lump* lump = (Lump*)lump_list->GetItemData(selection[a]);
			string name = wxGetTextFromUser(wxString::Format(_T("Enter new lump name for %s:"), lump->getName()), _T("Rename Lump"));

			if (!name.empty())
			{
				if (!lump->getParent()->zip)
				{
					name.UpperCase();
					name.Truncate(8);
				}

				lump->setName(name);
				lump_list->SetItemText(selection[a], name);
				lump->determineType();
			}
		}
	}
}

void WadPanel::importLump()
{
	if (selection.size() > 0)
	{
		for (int a = 0; a < selection.size(); a++)
		{
			Lump* lump = (Lump*)lump_list->GetItemData(selection[a]);
			string filename = wxFileSelector(wxString::Format(_T("Import %s"), lump->getName()), _T(""), _T(""), _T("*.*"),
												_T("All Files (*.*)|*.*"), wxOPEN|wxFILE_MUST_EXIST);

			if (filename != _T(""))
			{
				lump->loadFile(filename);
				lump->determineType();
				updateList(a);
			}
		}
	}
}

void WadPanel::exportImageLump(Lump* lump)
{
	wxString formats = _T("Raw Lump (*.lmp)|*.lmp|PNG (*.png)|*.png|Windows Bitmap (*.bmp)|*.bmp|All Files (*.*)|*.*");
	wxString deftype = _T("*.lmp");
	wxString fname = lump->getName();

	if (lump->getType() == LUMP_PNG)
		deftype = _T("*.png");

	wxString filename = wxFileSelector(wxString::Format(_T("Export %s"), lump->getName()),
										_T(""), fname, deftype, formats, wxSAVE|wxOVERWRITE_PROMPT);

	wxString ftype = filename.Right(3).Lower();

	if (!filename.empty())
		lump->dumpToFile(filename);
}

void WadPanel::exportLump()
{
	if (selection.size() > 0)
	{
		for (int a = 0; a < selection.size(); a++)
		{
			Lump* lump = (Lump*)lump_list->GetItemData(selection[a]);

			// Process gfx lumps differently
			if (lump->getType() == LUMP_GFX ||
				lump->getType() == LUMP_SPRITE ||
				lump->getType() == LUMP_PATCH ||
				lump->getType() == LUMP_FLAT ||
				lump->getType() == LUMP_PNG ||
				lump->getType() == LUMP_GFX2)
			{
				exportImageLump(lump);
				continue;
			}

			wxString filename = wxFileSelector(wxString::Format(_T("Export %s"), lump->getName()), _T(""),
												wxString::Format(_T("%s.lmp"), lump->getName()), _T("*.*"),
												_T("All Files (*.*)|*.*"), wxSAVE|wxOVERWRITE_PROMPT);

			if (!filename.empty())
				lump->dumpToFile(filename);
		}
	}
}

void WadPanel::exportLumpToWad()
{
	if (selection.size() == 0)
		return;

	wxString filename = wxFileSelector(_T("Export to Wad"), _T(""), _T(""), _T("*.wad"), _T("Doom Wad Files (*.wad)|*.wad"),
										wxSAVE|wxOVERWRITE_PROMPT);

	if (!filename.empty())
	{
		Wad newwad;
		newwad.path = filename;
		for (int a = 0; a < selection.size(); a++)
		{
			Lump* lump = (Lump*)lump_list->GetItemData(selection[a]);
			newwad.addLump(lump->getName(), -1);
			newwad.lastLump()->loadData(lump->getData(), lump->getSize());
		}

		newwad.save();
	}
}

void WadPanel::moveLump(bool up)
{
	updateSelection();

	if (up)
	{
		vector<wxListItem> items;

		if (selection[0] == 0)
			return;
		//else
		{
			for (int a = 0; a < selection.size(); a++)
			{
				// Swap lumps in the wad
				wadfile.swapLumps(selection[a], selection[a] - 1);

				// Update the lump list
				lump_list->SetItemData(selection[a], (long)wadfile.lumpAt(selection[a]));
				lump_list->SetItemData(selection[a]-1, (long)wadfile.lumpAt(selection[a]-1));
				updateList(selection[a]);
				updateList(selection[a]-1);
			}
		}

		// Modify selection
		for (int a = 0; a < wadfile.numLumps(); a++)
		{
			if (lump_list->GetItemState(a, wxLIST_STATE_SELECTED))
			{
				lump_list->SetItemState(a, 0, wxLIST_STATE_SELECTED|wxLIST_STATE_FOCUSED);
				if (a > 0)
					lump_list->SetItemState(a-1, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
			}
		}
	}
	else
	{
		if (selection.back() == wadfile.numLumps() - 1)
			return;
		else
		{
			for (int a = selection.size() - 1; a >= 0; a--)
			{
				// Swap lumps in the wad
				wadfile.swapLumps(selection[a], selection[a] + 1);

				// Update the lump list
				lump_list->SetItemData(selection[a], (long)wadfile.lumpAt(selection[a]));
				lump_list->SetItemData(selection[a]+1, (long)wadfile.lumpAt(selection[a]+1));
				updateList(selection[a]);
				updateList(selection[a]+1);
			}
		}

		// Modify selection
		for (int a = wadfile.numLumps() - 1; a >= 0; a--)
		{
			if (lump_list->GetItemState(a, wxLIST_STATE_SELECTED))
			{
				lump_list->SetItemState(a, 0, wxLIST_STATE_SELECTED|wxLIST_STATE_FOCUSED);
				if (a < wadfile.numLumps() - 1)
					lump_list->SetItemState(a+1, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
			}
		}
	}
}


BEGIN_EVENT_TABLE(WadPanel, wxPanel)
	EVT_LIST_ITEM_FOCUSED(WP_LIST_LUMPS, WadPanel::onLumpListChange)
	EVT_LIST_ITEM_SELECTED(WP_LIST_LUMPS, WadPanel::onLumpListSelect)
	EVT_LIST_ITEM_DESELECTED(WP_LIST_LUMPS, WadPanel::onLumpListDeSelect)
	EVT_LIST_COL_END_DRAG(WP_LIST_LUMPS, WadPanel::onLumpListColResize)
	EVT_LIST_END_LABEL_EDIT(WP_LIST_LUMPS, WadPanel::onLumpListEditName)
END_EVENT_TABLE()

void WadPanel::onLumpListColResize(wxListEvent &event)
{
	int width = 0;
	for (int a = 0; a < lump_list->GetColumnCount(); a++)
		width += lump_list->GetColumnWidth(a);

	lump_list->SetSizeHints(width + 20, -1, width + 20, -1);

	Layout();
}

void WadPanel::onLumpListSelect(wxListEvent &event)
{
	selection.push_back(event.GetIndex());
}

void WadPanel::onLumpListDeSelect(wxListEvent &event)
{
	for (int a = 0; a < selection.size(); a++)
	{
		if (selection[a] == event.GetIndex())
		{
			selection[a] = selection.back();
			selection.pop_back();
			return;
		}
	}
}

void WadPanel::onLumpListChange(wxListEvent &event)
{
	// Multiple items selected
	if (selection.size() > 1)
	{
		lump_area_sizer->Detach(lump_area);
		delete(lump_area);
		lump_area = new LumpArea(this);
		lump_area_sizer->Add(lump_area, 1, wxEXPAND);

		int totalsize = 0;

		for (int a = 0; a < selection.size(); a++)
		{
			Lump* lump = (Lump*)lump_list->GetItemData(selection[a]);
			totalsize += lump->getSize();
		}

		main_window->SetStatusText(wxString::Format(_T("Size: %d"), totalsize), 1);

		main_window->SetStatusText(_T(""), 2);
		lump_area->Show();
		lump_area_sizer->Layout();

		return;
	}

	Lump* lump = (Lump*)lump_list->GetItemData(event.GetIndex());

	if (!lump)
		return;

	lump->determineType();
	lump_area->checkSave();
	updateList(event.GetIndex());

	if (lump->getType() == LUMP_GFX ||
		lump->getType() == LUMP_SPRITE ||
		lump->getType() == LUMP_PATCH ||
		lump->getType() == LUMP_FLAT ||
		lump->getType() == LUMP_PNG ||
		lump->getType() == LUMP_GFX2)
	{
		if (lump_area->type != LAREA_GFX)
		{
			lump_area_sizer->Detach(lump_area);
			delete(lump_area);
			lump_area = new GfxLumpArea(this, palette);
			lump_area_sizer->Add(lump_area, 1, wxEXPAND);
		}
	}
	else if (lump->getType() == LUMP_TEXT || lump->getType() == LUMP_UNKNOWN || lump->getType() == LUMP_MARKER)
	{
		if (lump_area->type != LAREA_TEXT)
		{
			lump_area_sizer->Detach(lump_area);
			delete(lump_area);
			lump_area = new TextLumpArea(this);
			lump_area_sizer->Add(lump_area, 1, wxEXPAND);
		}
	}
	else if (lump->getType() == LUMP_TEXTURES)
	{
		if (lump_area->type != LAREA_TEXTURES)
		{
			lump_area_sizer->Detach(lump_area);
			delete(lump_area);
			lump_area = new TextureLumpArea(this, palette, &wadfile);
			lump_area_sizer->Add(lump_area, 1, wxEXPAND);
		}
	}
	else if (lump->getType() == LUMP_MAP)
	{
		if (lump_area->type != LAREA_MAP)
		{
			lump_area_sizer->Detach(lump_area);
			delete(lump_area);
			lump_area = new MapLumpArea(this, &wadfile);
			lump_area_sizer->Add(lump_area, 1, wxEXPAND);
		}
	}
	else
	{
		if (lump_area->type != LAREA_NONE)
		{
			lump_area_sizer->Detach(lump_area);
			delete(lump_area);
			lump_area = new LumpArea(this);
			lump_area_sizer->Add(lump_area, 1, wxEXPAND);
		}
	}

	lump_area->loadLump(lump);
	main_window->SetStatusText(wxString::Format(_T("Size: %d"), lump->getSize()), 1);
	main_window->SetStatusText(wxString::Format(_T("Offset: %d"), lump->getOffset()), 2);
	lump_area->Show();
	lump_area->Update();
	lump_area_sizer->Layout();

	Layout();
}

void WadPanel::onLumpListEditName(wxListEvent &event)
{
	/*
	wxMessageBox(event.GetText());

	Lump* lump = (Lump*)lump_list->GetItemData(event.GetIndex());

	if (lump)
		lump->setName(event.GetLabel());
		*/
}
